﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis {
  [StorableClass]
  [Item("AverageThresholdCalculator", "")]
  public class AverageThresholdCalculator : DiscriminantClassificationWeightCalculator {

    public AverageThresholdCalculator()
      : base() {
    }
    [StorableConstructor]
    protected AverageThresholdCalculator(bool deserializing) : base(deserializing) { }
    protected AverageThresholdCalculator(AverageThresholdCalculator original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new AverageThresholdCalculator(this, cloner);
    }

    protected double[] threshold;
    protected double[] classValues;

    protected override IEnumerable<double> DiscriminantCalculateWeights(IEnumerable<IDiscriminantFunctionClassificationSolution> discriminantSolutions) {
      classValues = discriminantSolutions.First().Model.ClassValues.ToArray();
      var modelThresholds = discriminantSolutions.Select(x => x.Model.Thresholds.ToArray());
      threshold = new double[modelThresholds.First().GetLength(0)];
      for (int i = 0; i < modelThresholds.First().GetLength(0); i++) {
        threshold[i] = modelThresholds.Select(x => x[i]).Average();
      }
      return Enumerable.Repeat<double>(1, discriminantSolutions.Count());
    }

    protected override double GetDiscriminantConfidence(IEnumerable<IDiscriminantFunctionClassificationSolution> solutions, int index, double estimatedClassValue, CheckPoint handler) {
      Dataset dataset = solutions.First().ProblemData.Dataset;
      IList<double> values = solutions.Where(s => handler(s.ProblemData, index)).Select(s => s.Model.GetEstimatedValues(dataset, Enumerable.Repeat(index, 1)).First()).ToList();
      if (values.Count <= 0)
        return double.NaN;
      double avg = values.Average();
      return GetAverageConfidence(avg, estimatedClassValue);
    }

    public override IEnumerable<double> GetDiscriminantConfidence(IEnumerable<IDiscriminantFunctionClassificationSolution> solutions, IEnumerable<int> indices, IEnumerable<double> estimatedClassValue, CheckPoint handler) {
      Dataset dataset = solutions.First().ProblemData.Dataset;
      List<int> indicesList = indices.ToList();
      var solValues = solutions.ToDictionary(x => x, x => x.Model.GetEstimatedValues(dataset, indicesList).ToArray());
      double[] confidences = new double[indices.Count()];
      double[] estimatedClassValueArr = estimatedClassValue.ToArray();

      for (int i = 0; i < indicesList.Count; i++) {
        var values = solValues.Where(x => handler(x.Key.ProblemData, indicesList[i])).Select(x => x.Value[i]);
        if (values.Count() <= 0) {
          confidences[i] = double.NaN;
        } else {
          double avg = values.Average();
          confidences[i] = GetAverageConfidence(avg, estimatedClassValueArr[i]);
        }
      }

      return confidences;
    }

    protected double GetAverageConfidence(double avg, double estimatedClassValue) {
      for (int i = 0; i < classValues.Length; i++) {
        if (estimatedClassValue.Equals(classValues[i])) {
          //special case: avgerage is higher than value of highest class
          if (i == classValues.Length - 1 && avg >= estimatedClassValue) {
            return 1;
          }
          //special case: average is lower than value of lowest class
          if (i == 0 && avg < estimatedClassValue) {
            return 1;
          }
          //special case: average is not between threshold of estimated class value
          if ((i < classValues.Length - 1 && avg >= threshold[i + 1]) || avg <= threshold[i]) {
            return 0;
          }

          double thresholdToClassDistance, thresholdToAverageValueDistance;
          if (avg >= classValues[i]) {
            thresholdToClassDistance = threshold[i + 1] - classValues[i];
            thresholdToAverageValueDistance = threshold[i + 1] - avg;
          } else {
            thresholdToClassDistance = classValues[i] - threshold[i];
            thresholdToAverageValueDistance = avg - threshold[i];
          }
          return (1 / thresholdToClassDistance) * thresholdToAverageValueDistance;
        }
      }
      return double.NaN;
    }
  }
}
